import json
from typing import Dict, List

from aiohttp import ClientSession

from omni_reports.client import ReportClient
from omni_reports.client.errors import ReportResponseError
from omni_reports.client.models import ReportPredicate, ReportDefinitionDateRange
from omni_reports.client.types import ReportType
from omni_reports.facebook_reports.fields import FacebookReportField, FACEBOOK_OPERATORS_MAP
from omni_reports.facebook_reports.settings import (
    FACEBOOK_BASE_API, FACEBOOK_VERSION_API,
    REPORT_ENDPOINT_PATTERN
)


def get_base_url():
    return f"{FACEBOOK_BASE_API}/{FACEBOOK_VERSION_API}"


class FacebookAdsReportType(ReportType):
    REPORT_TYPE = None
    FIELD_BEGIN_SEPARATOR = "{"
    FIELD_END_SEPARATOR = "}"

    async def resolve(self, fields, predicates, report_definition, context, client: ReportClient):
        if not self.REPORT_TYPE:
            raise AttributeError('The attribute "facebook_report_type" is required')

        if client.session is None:
            raise AttributeError('Client session cannot be null.')

        session = client.session
        date_range: ReportDefinitionDateRange = report_definition.selector.date_range

        self._validate_predicates(predicates, fields, date_range)

        formatted_fields = self._get_fields(fields)
        predicates = self._get_predicates(predicates)

        url = self._build_request_url(context, formatted_fields, predicates, date_range)
        records = await request(session, url)

        return self._resolve_fields_on_records(records, fields)

    def _build_request_url(self, context, fields, predicates, date_range):

        token = context.get("FACEBOOK_TOKEN")
        network_id = context.get("FACEBOOK_NETWORK_ID")

        endpoint_pattern = REPORT_ENDPOINT_PATTERN[self.REPORT_TYPE] or ""
        endpoint = self._compose_endpoint_with_predicates(endpoint_pattern, {
            "network_id": network_id
        })
        date_range = self._compose_date_range(date_range)
        predicates = self._compose_predicates(predicates)

        return f"{get_base_url()}/{endpoint}?fields={fields}" \
               f"&access_token={token}&{date_range}" \
               f"&filtering={predicates}"

    @staticmethod
    def _validate_predicates(
            predicates: Dict[str, ReportPredicate],
            fields: dict,
            date_range: ReportDefinitionDateRange
    ):
        for predicate in predicates.values():
            if predicate.field.target_group_name:
                raise AttributeError(f"Fields with target group is not accepted yet. "
                                     f"Invalid field: {predicate.field.target_group_name}"

                                     f" - {predicate.field.target_name}")

        if date_range:
            has_date_field = bool([
                field_name
                for field_name, _ in fields.items()
                if field_name == 'date'
            ])
            if date_range.time_increment != 1 and has_date_field:
                raise AttributeError(
                    "You can only date field with time_increment equals to 1. For"
                    "other time_increment values use date_start and date_end fields."
                )

    @staticmethod
    def _get_predicates(predicates: Dict[str, ReportPredicate]) -> List[dict]:
        formatted_predicates = []
        for predicate in predicates.values():
            operator = FACEBOOK_OPERATORS_MAP[predicate.operator]

            formatted_predicates.append({
                "field": predicate.field.target_name,
                "operator": operator,
                "value": predicate.values[0] if operator == "EQUAL" else predicate.values
            })

        return formatted_predicates

    @staticmethod
    def _resolve_fields_on_records(records, fields) -> List[dict]:
        if not isinstance(records, dict):
            raise ValueError(f'Invalid records response for {records}')
        if 'error' in records:
            raise ReportResponseError(
                f"\n\nFacebook Api response error:\n{records.get('error')}"
            )

        only_records = records.get("data") or []
        only_records = only_records if isinstance(only_records, list) else list(only_records)

        updated_records = []

        for record in only_records:
            new_record = {}
            for field_name, field in fields.items():
                try:
                    new_record[field_name] = record[field.target_name]
                except KeyError as error:
                    raise KeyError(f"{error}\nField '{field_name}' not mapped on record {record}.")

                updated_records.append(new_record)

        return updated_records

    @staticmethod
    def _compose_predicates(predicates: dict) -> str:
        return json.dumps(predicates) if predicates else ""

    @staticmethod
    def _compose_date_range(date_range: ReportDefinitionDateRange) -> str:
        if not date_range:
            return ""

        start_range = "{start_object}'since':'{since}'".format(
            start_object="{",
            since=date_range.start,
        )
        end_range = "'until': '{until}'{end_object}".format(
            until=date_range.end,
            end_object="}"
        )

        time_range = f"time_range={start_range},{end_range}"
        time_increment = f"time_increment={date_range.time_increment}"

        return f"{time_range}&{time_increment}"

    def _get_fields(self, fields: Dict[str, FacebookReportField]):
        group_fields = {
            field.target_group_name
            for _, field in fields.items()
            if field.target_group_name
        }

        grouped_fields = {}
        for group_field in group_fields:
            filtered_fields = (
                field.target_name
                for _, field in fields.items()
                if group_field == field.target_group_name
            )
            grouped_fields[group_field] = \
                group_field \
                + self.FIELD_BEGIN_SEPARATOR \
                + ",".join(filtered_fields) \
                + self.FIELD_END_SEPARATOR

        non_object_fields = [
            field.target_name
            for _, field in fields.items()
            if not field.target_group_name
        ]
        all_fields = ",".join([*non_object_fields, *grouped_fields.values()])

        return all_fields

    @staticmethod
    def _compose_endpoint_with_predicates(pattern: str, predicates: dict):
        return pattern.format(**predicates)

    @staticmethod
    def _get_predicates_not_used_on_endpoint_pattern(pattern: str, predicates: dict):
        return {
            predicate_field: predicates[predicate_field]
            for predicate_field in predicates.keys()
            if predicate_field not in pattern
        }


async def request(session: ClientSession, url: str):
    async with session.get(url) as response:
        return await response.json()



import inspect

from omni_reports.client.errors import ReportTypeFieldNotFoundError, InvalidReportRecordError
from omni_reports.client.fields import ReportFieldABC
from omni_reports.client.models import ReportPredicate
from omni_reports.client.utils import is_instance_or_subclass


def _get_fields(attrs, field_class, pop=False):
    fields = [
        (field_name, field_value)
        for field_name, field_value in attrs.items()
        if is_instance_or_subclass(field_value, field_class)
    ]
    if pop:
        for field_name, _ in fields:
            del attrs[field_name]
    return fields


def _get_fields_by_mro(klass, field_class):
    mro = inspect.getmro(klass)
    return sum(
        (
            _get_fields(
                getattr(base, "_declared_fields", base.__dict__),
                field_class,
            )
            for base in mro[:0:-1]
        ),
        [],
    )


class ReportTypeMeta(type):
    def __new__(mcs, name, bases, attrs):
        cls_fields = _get_fields(attrs, ReportFieldABC, True)
        klass = super().__new__(mcs, name, bases, attrs)
        inherited_fields = _get_fields_by_mro(klass, ReportFieldABC)

        meta = klass.Meta
        klass.opts = klass.OPTIONS_CLASS(meta)

        klass._declared_fields = mcs.get_declared_fields(
            klass=klass,
            cls_fields=cls_fields,
            inherited_fields=inherited_fields,
        )

        return klass

    @classmethod
    def get_declared_fields(
            mcs,
            klass,
            cls_fields,
            inherited_fields,
    ):
        return dict(inherited_fields + cls_fields)


class ReportTypeOptions:
    def __init__(self, meta):
        self.meta = meta


class ReportTypeABC:
    def execute(self, report_definition, context, client):
        raise NotImplementedError

    def resolve(self, fields, predicates, report_definition, context, client):
        raise NotImplementedError


class ReportType(ReportTypeABC, metaclass=ReportTypeMeta):
    OPTIONS_CLASS = ReportTypeOptions

    _declared_fields = {}

    opts = None  # type: ReportTypeOptions

    class Meta:
        pass

    async def execute(self, report_definition, context, client):
        fields = self._extract_fields(report_definition)
        predicates = self._extract_predicates(report_definition)

        records = await self.resolve(
            fields,
            predicates,
            report_definition,
            context,
            client,
        )

        return self._filter_fields_in_records(records, fields)

    async def resolve(self, fields, predicates, report_definition, context, client):
        raise NotImplementedError

    def _extract_predicates(self, report_definition):
        predicates = {}

        for rd_predicate in report_definition.selector.predicates:
            field = self._declared_fields.get(rd_predicate.field)
            if not field:
                raise ReportTypeFieldNotFoundError(rd_predicate.field)
            predicates[rd_predicate.field] = ReportPredicate(
                field=field,
                operator=rd_predicate.operator,
                values=rd_predicate.values
            )

        return predicates

    def _extract_fields(self, report_definition):
        fields = {}
        for rd_field in report_definition.selector.fields:
            field = self._declared_fields.get(rd_field)
            if not field:
                raise ReportTypeFieldNotFoundError(rd_field)
            fields[rd_field] = field
        return fields

    @staticmethod
    def _filter_fields_in_records(records, fields):
        new_records = []
        for record in records:
            new_record = {}
            for field_name, field in fields.items():
                if field_name not in record:
                    raise InvalidReportRecordError(record, f"Field with name \"{field_name}\" not found on record")
                new_record[field_name] = record[field_name]
            new_records.append(new_record)
        return new_records



from omni_reports.client import ReportTypeResolverBuilder

from omni_reports.google_reports.types import (
    GoogleAdsAccountPerformanceReportType,
    GoogleAdsKeywordsPerformanceReportType
)

GoogleAdsReportTypeResolver = ReportTypeResolverBuilder() \
    .add_type('GOOGLE_ADS_ACCOUNT_PERFORMANCE_REPORT', GoogleAdsAccountPerformanceReportType()) \
    .add_type('GOOGLE_ADS_KEYWORDS_PERFORMANCE_REPORT', GoogleAdsKeywordsPerformanceReportType()) \
    .build()



import asyncio
import os
from pprint import pprint

from dotenv import load_dotenv

from omni_reports.client import ReportClient
from omni_reports.google_reports import GoogleAdsReportTypeResolver

load_dotenv()


async def request():
    client = ReportClient(GoogleAdsReportTypeResolver, {
        'GOOGLE_ADS_CLIENT_ID': os.getenv('GOOGLE_ADS_CLIENT_ID'),
        'GOOGLE_ADS_CLIENT_SECRET': os.getenv('GOOGLE_ADS_CLIENT_SECRET'),
        'GOOGLE_ADS_DEVELOPER_TOKEN': os.getenv('GOOGLE_ADS_DEVELOPER_TOKEN'),
        'GOOGLE_ADS_USER_AGENT': os.getenv('GOOGLE_ADS_USER_AGENT'),
        'GOOGLE_ADS_CUSTOMER_ID': os.getenv('GOOGLE_ADS_CUSTOMER_ID'),
        'GOOGLE_ADS_REFRESH_TOKEN': os.getenv('GOOGLE_ADS_REFRESH_TOKEN'),
    })

    result = await client.execute_report({
        "report_type": "GOOGLE_ADS_KEYWORDS_PERFORMANCE_REPORT",
        "report_name": "this_month_google_ads_keywords_performance",
        "selector": {
            "fields": ["id", "criteria", "cost"],
            "date_range": {
                "start": "2020-02-15",
                "end": "2020-02-16"
            },
            "predicates": [
                {
                    "field": "has_quality_score",
                    "operator": "equals",
                    "values": ["TRUE"]
                }
            ]
        }
    })

    pprint(result)


async def main():
    await request()


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())



from omni_reports.client.fields import AttributeReportField, MetricReportField, SegmentReportField
from omni_reports.google_reports.base import GoogleAdsReportType

BOOLEAN_VALUES = {
    'true': True,
    'false': False
}


def to_bool(val):
    return BOOLEAN_VALUES.get(val)


class GoogleAdsAccountPerformanceReportType(GoogleAdsReportType):
    REPORT_TYPE = "ACCOUNT_PERFORMANCE_REPORT"

    account_id = AttributeReportField(target_name="ExternalCustomerId", display_name="Customer ID")
    day = SegmentReportField(target_name="Date", display_name="Day")
    cost = MetricReportField(target_name="Cost", display_name="Cost", map=lambda val: float(val) / 1e6)
    currency = AttributeReportField(target_name="AccountCurrencyCode", display_name="Currency")
    conversions = MetricReportField(target_name="Conversions", display_name="Conversions", type=float)
    cost_per_conversion = MetricReportField(target_name="CostPerConversion", display_name="Cost / conv.",
                                            map=lambda val: float(val) / 1e6)


class GoogleAdsKeywordsPerformanceReportType(GoogleAdsReportType):
    REPORT_TYPE = "KEYWORDS_PERFORMANCE_REPORT"

    id = AttributeReportField(target_name="Id", display_name="Keyword ID")
    criteria = AttributeReportField(target_name="Criteria", display_name="Keyword")
    ad_group_id = AttributeReportField(target_name="AdGroupId", display_name="Ad group ID")
    ad_group_name = AttributeReportField(target_name="AdGroupName", display_name="Ad group")
    ad_group_status = AttributeReportField(target_name="AdGroupStatus", display_name="Ad group state")
    campaign_id = AttributeReportField(target_name="CampaignId", display_name="Campaign ID")
    campaign_name = AttributeReportField(target_name="CampaignName", display_name="Campaign")
    campaign_status = AttributeReportField(target_name="CampaignStatus", display_name="Campaign state")
    is_negative = AttributeReportField(target_name="IsNegative", display_name="Is negative", map=to_bool)
    has_quality_score = AttributeReportField(target_name="HasQualityScore",
                                             display_name="Has Quality Score",
                                             map=to_bool)
    quality_score = AttributeReportField(target_name="QualityScore", display_name="Quality score")

    cost = MetricReportField(target_name="Cost", display_name="Cost")



from omni_reports.client import ReportTypeResolverBuilder
from omni_reports.facebook_reports.types import FacebookAdsAccountPerformanceReportType

FacebookAdsReportTypeResolver = ReportTypeResolverBuilder() \
    .add_type('FACEBOOK_ADS_ACCOUNT_PERFORMANCE_REPORT', FacebookAdsAccountPerformanceReportType()) \
    .build()




class ReportTypeNotFoundError(RuntimeError):
    def __init__(self, type_name):
        super(ReportTypeNotFoundError, self).__init__(
            f"No report type found with name \"{type_name}\"."
        )


class ReportTypeFieldNotFoundError(RuntimeError):
    def __init__(self, field_name):
        super(ReportTypeFieldNotFoundError, self).__init__(
            f"No report type field found with name \"{field_name}\"."
        )


class InvalidReportRecordError(RuntimeError):
    def __init__(self, record, message):
        super(InvalidReportRecordError, self).__init__(message)
        self.record = record


class InvalidReportDefinitionError(RuntimeError):
    def __init__(self, report_definition, message):
        super().__init__(message)


class ReportResponseError(RuntimeError):
    def __init__(self, message):
        super().__init__(message)



import csv
from io import TextIOWrapper

from googleads import oauth2, adwords, common

from omni_reports.client import ReportClient
from omni_reports.client.types import ReportType
from omni_reports.client.utils import async_wrap

GOOGLE_OPERATORS_MAP = {
    'equals': 'EQUALS',
    'not_equals': 'NOT_EQUALS',
    'in': 'IN',
    'not_in': 'NOT_IN',
}


class GoogleAdsReportType(ReportType):
    REPORT_TYPE = None

    async def resolve(self, fields, predicates, report_definition, context, client: ReportClient):
        google_report_type = self.REPORT_TYPE
        if not google_report_type:
            raise AttributeError('The attribute "google_report_type" is required')

        gads_report_definition = self.__convert_to_google_ads_report_definition(
            fields,
            predicates,
            report_definition,
            google_report_type
        )

        gads_client = self.__create_client(context)
        async_func = async_wrap(self.retrieve_report)

        return await async_func(
            fields,
            gads_client,
            gads_report_definition,
        )

    def retrieve_report(self, fields, gads_client, gads_report_definition):
        gads_report_stream = self.__download_report_as_stream(gads_client, gads_report_definition)
        gads_report_stream_wrapper = TextIOWrapper(gads_report_stream, encoding='utf-8')
        gads_report_reader = csv.DictReader(gads_report_stream_wrapper)

        return self.__convert_from_google_ads_records(gads_report_reader, fields)

    @staticmethod
    def __convert_to_google_ads_report_definition(fields, predicates, report_definition, google_report_type):
        gads_predicates = []
        for predicate_name, predicate in predicates.items():
            gads_predicates.append({
                'field': predicate.field.target_name,
                'operator': GOOGLE_OPERATORS_MAP[predicate.operator],
                'values': predicate.values
            })
        return {
            'reportType': google_report_type,
            'reportName': report_definition.report_name,
            'dateRangeType': 'CUSTOM_DATE',
            'downloadFormat': 'CSV',
            'selector': {
                'fields': [field.target_name or field_name for field_name, field in fields.items()],
                'predicates': gads_predicates,
                'dateRange': {
                    'min': str(report_definition.selector.date_range.start),
                    'max': str(report_definition.selector.date_range.end)
                }
            }
        }

    @staticmethod
    def __convert_from_google_ads_records(records, fields):
        converted_records = []
        for record in records:
            converted_record = dict()
            for column, value in record.items():
                field = next(
                    (name for name, field in fields.items() if field.extra.get('display_name') == column), None)
                if not field:
                    raise AttributeError(f'Field with name "{column}" not found')

                field_definition = fields[field]
                if 'map' in field_definition.extra:
                    value_map = field_definition.extra.get('map')
                    value = value_map(value)
                if 'type' in field_definition.extra:
                    value_type = field_definition.extra.get('type')
                    value = value_type(value)

                converted_record[field] = value
            converted_records.append(converted_record)
        return converted_records

    @staticmethod
    def __create_client(context):
        client_id = context.get('GOOGLE_ADS_CLIENT_ID')
        client_secret = context.get('GOOGLE_ADS_CLIENT_SECRET')
        refresh_token = context.get('GOOGLE_ADS_REFRESH_TOKEN')
        developer_token = context.get('GOOGLE_ADS_DEVELOPER_TOKEN')
        user_agent = context.get('GOOGLE_ADS_USER_AGENT')
        customer_id = context.get('GOOGLE_ADS_CUSTOMER_ID')

        oauth2_client = oauth2.GoogleRefreshTokenClient(
            client_id,
            client_secret,
            refresh_token)

        ads_client = adwords.AdWordsClient(
            developer_token,
            oauth2_client,
            user_agent,
            client_customer_id=customer_id)
        ads_client.cache = common.ZeepServiceProxy.NO_CACHE

        return ads_client

    @staticmethod
    def __download_report_as_stream(client, report):
        report_downloader = client.GetReportDownloader(version='v201809')

        return report_downloader.DownloadReportAsStream(
            report, skip_report_header=True, skip_column_header=False,
            skip_report_summary=True, include_zero_impressions=False)



from marshmallow import Schema, post_load, validate
from marshmallow import fields as ma_fields

from omni_reports.client import models


class ReportDefinitionPredicateSchema(Schema):
    field = ma_fields.Str(required=True)
    operator = ma_fields.Str(required=True)
    values = ma_fields.List(ma_fields.Str(), validate=validate.Length(min=1))

    @post_load
    def create(self, data, **kwargs):
        return models.ReportDefinitionPredicate(**data)


class ReportDefinitionDateRangeSchema(Schema):
    start = ma_fields.Date(format="%Y-%m-%d", required=True)
    end = ma_fields.Date(format="%Y-%m-%d", required=True)
    time_increment = ma_fields.Integer(required=False)

    @post_load
    def create(self, data, **kwargs):
        return models.ReportDefinitionDateRange(**data)


class ReportDefinitionSelectorSchema(Schema):
    fields = ma_fields.List(ma_fields.Str(), required=True, validate=validate.Length(min=1))
    predicates = ma_fields.List(ma_fields.Nested(ReportDefinitionPredicateSchema()), required=False)
    date_range = ma_fields.Nested(ReportDefinitionDateRangeSchema(), required=True)

    @post_load
    def create(self, data, **kwargs):
        return models.ReportDefinitionSelector(**data)


class ReportDefinitionSchema(Schema):
    report_type = ma_fields.Str(required=True)
    report_name = ma_fields.Str(required=True)
    selector = ma_fields.Nested(ReportDefinitionSelectorSchema(), required=True)

    @post_load
    def create(self, data, **kwargs):
        return models.ReportDefinition(**data)


class ReportSchema(Schema):
    report_definition = ma_fields.Nested(ReportDefinitionSchema(), required=True)
    records = ma_fields.List(ma_fields.Dict(), required=True)

    @post_load
    def create(self, data, **kwargs):
        return models.Report(**data)



from omni_reports.facebook_reports.base import FacebookAdsReportType

from omni_reports.facebook_reports.fields import (
    FacebookAttributeReportField, FacebookMetricReportField
)

BOOLEAN_VALUES = {
    'true': True,
    'false': False
}


def to_bool(val):
    return BOOLEAN_VALUES.get(val)


class FacebookAdsAccountPerformanceReportType(FacebookAdsReportType):
    REPORT_TYPE = "FACEBOOK_ACCOUNT_REPORT"

    customer_id = FacebookAttributeReportField(target_name="account_id")
    date = FacebookAttributeReportField(target_name="date_start")
    date_start = FacebookAttributeReportField(target_name="date_start")
    date_end = FacebookAttributeReportField(target_name="date_end")
    currency = FacebookMetricReportField(target_name="account_currency")
    cost = FacebookMetricReportField(target_name="spend")
    ad_set_id = FacebookMetricReportField(target_name="id", target_group_name="adsets")



import asyncio
import os
from pprint import pprint

from aiohttp import ClientSession
from dotenv import load_dotenv

from omni_reports.client import ReportClient
from omni_reports.facebook_reports import FacebookAdsReportTypeResolver

load_dotenv()


async def request():
    async with ClientSession() as session:
        client = ReportClient(FacebookAdsReportTypeResolver, {
            'FACEBOOK_TOKEN': os.getenv('FACEBOOK_TOKEN'),
            'FACEBOOK_NETWORK_ID': os.getenv('FACEBOOK_NETWORK_ID'),
        }, session=session)

        result = await client.execute_report({
            "report_type": "FACEBOOK_ADS_ACCOUNT_PERFORMANCE_REPORT",
            "report_name": "test",
            "selector": {
                "fields": ["cost", "date", "currency"],
                "date_range": {
                    "start": "2000-02-15",
                    "end": "2000-02-16",
                },
                "predicates": [
                    {
                        "field": "cost",
                        "operator": "equals",
                        "values": ["1396.77"]
                    }
                ]
            }
        })

        pprint(result)


async def main():
    await request()


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())



from typing import Optional, Dict

from aiohttp import ClientSession

from omni_reports.client.models import Report
from omni_reports.client.resolvers import ReportTypeResolverABC
from omni_reports.client.schemas import ReportDefinitionSchema, ReportSchema


class ReportClient:
    __report_definition_schema = ReportDefinitionSchema()
    __report_schema = ReportSchema()

    session: Optional[ClientSession]

    def __init__(
            self, resolver: ReportTypeResolverABC,
            context: Optional[Dict] = None,
            session: Optional[ClientSession] = None
    ):
        self.resolver = resolver
        self.context = context or dict()

        self.session = session

    async def execute_report(self, report_definition, context=None):
        context = context or dict()
        context = {**self.context, **context}

        report_def = self.__report_definition_schema.load(report_definition)
        report_records = await self.resolver.process(report_def, context, self)
        report = Report(
            report_definition=report_def,
            records=report_records,
        )

        return self.__report_schema.dump(report)



from omni_reports.client.fields import (
    ReportField, AttributeReportField,
    SegmentReportField, MetricReportField
)

FACEBOOK_OPERATORS_MAP = {
    'equals': 'EQUAL',
    'not_equals': 'NOT_EQUALS',
    'in': 'IN',
    'not_in': 'NOT_IN',
}


class FacebookReportField(ReportField):
    def __init__(self, *args, target_name, target_group_name=None, **extra):
        self.target_group_name = target_group_name

        super().__init__(*args, target_name=target_name, **extra)


class FacebookAttributeReportField(FacebookReportField, AttributeReportField):
    pass


class FacebookSegmentReportField(FacebookReportField, SegmentReportField):
    pass


class FacebookMetricReportField(FacebookReportField, MetricReportField):
    pass



from omni_reports.client.errors import ReportTypeNotFoundError


class ReportTypeResolverABC:
    async def process(self, report_definition, context, client):
        raise NotImplementedError()


class ReportTypeResolver(ReportTypeResolverABC):
    def __init__(self, types=None):
        self.types = types or {}

    async def process(self, report_definition, context, client):
        type_config = self.types.get(report_definition.report_type)
        if not type_config:
            raise ReportTypeNotFoundError(report_definition)
        return await type_config.execute(report_definition, context, client)


class ReportTypeResolverBuilder:
    __types = {}

    def extend(self, resolver: ReportTypeResolver):
        self.__types.update(resolver.types)
        return self

    def add_type(self, report_type, report_type_config):
        self.__types[report_type] = report_type_config
        return self

    def build(self) -> ReportTypeResolverABC:
        return ReportTypeResolver(
            types=self.__types,
        )



from asyncio import get_running_loop
from functools import wraps, partial


def is_instance_or_subclass(val, class_) -> bool:
    """Return True if ``val`` is either a subclass or instance of ``class_``."""
    try:
        return issubclass(val, class_)
    except TypeError:
        return isinstance(val, class_)


def async_wrap(func):
    @wraps(func)
    async def run(*args, loop=None, executor=None, **kwargs):
        event_loop = loop or get_running_loop()

        p_func = partial(func, *args, **kwargs)
        return await event_loop.run_in_executor(executor, p_func)

    return run



from datetime import date
from typing import Dict, List

from omni_reports.client.fields import ReportField


class ReportPredicate:
    def __init__(self, field: ReportField = None, operator: str = None, values: List[str] = None):
        self.field = field
        self.operator = operator
        self.values = values

    def __str__(self):
        return f"<ReportPredicate field={self.field} operator={self.operator} values={self.values}>"

    def __repr__(self):
        return self.__str__()


class ReportDefinitionPredicate:
    def __init__(self, field: str = None, operator: str = None, values: List[str] = None):
        self.field = field
        self.operator = operator
        self.values = values

    def __str__(self):
        return f"<ReportDefinitionPredicate field={self.field} operator={self.operator} values={self.values}>"

    def __repr__(self):
        return self.__str__()


class ReportDefinitionDateRange:
    def __init__(self, start: date = None, end: date = None, time_increment: int = 1):
        self.start = start
        self.end = end
        self.time_increment = time_increment or 1

    def __str__(self):
        return f"<ReportDefinitionDateRange start={self.start} end={self.end}>"

    def __repr__(self):
        return self.__str__()

    def __bool__(self):
        return bool(self.start and self.end and self.time_increment)


class ReportDefinitionSelector:
    def __init__(self,
                 fields: List[str] = None,
                 predicates: List[ReportDefinitionPredicate] = None,
                 date_range: ReportDefinitionDateRange = None):
        self.fields = fields
        self.predicates = predicates or list()
        self.date_range = date_range

    def __str__(self):
        return f"<ReportDefinitionSelector fields={self.fields}>"

    def __repr__(self):
        return self.__str__()


class ReportDefinition:
    def __init__(self, report_type: str = None, report_name: str = None, selector: ReportDefinitionSelector = None):
        self.report_type = report_type
        self.report_name = report_name
        self.selector = selector

    def __str__(self):
        return f"<ReportDefinition report_type={self.report_type} report_name={self.report_name}>"

    def __repr__(self):
        return self.__str__()


class Report:
    def __init__(self, report_definition: ReportDefinition = None, records: List[Dict] = None):
        self.report_definition = report_definition
        self.records = records

    def __str__(self):
        return f"<Report " \
               f"report_type={self.report_definition.report_type} " \
               f"report_name={self.report_definition.report_name}>"

    def __repr__(self):
        return self.__str__()



class ReportFieldABC:
    pass


class ReportField(ReportFieldABC):
    def __init__(self, name=None, behavior=None, target_name=None, **extra):
        self.name = name
        self.behavior = behavior
        self.extra = extra

        # For predicate purposes
        self.target_name = target_name


class AttributeReportField(ReportField):
    def __init__(self, name=None, target_name=None, **extra):
        super(AttributeReportField, self).__init__(
            behavior="attribute",
            name=name,
            target_name=target_name,
            **extra
        )


class SegmentReportField(ReportField):
    def __init__(self, name=None, target_name=None, **extra):
        super(SegmentReportField, self).__init__(
            behavior="segment",
            name=name,
            target_name=target_name,
            **extra
        )


class MetricReportField(ReportField):
    def __init__(self, name=None, target_name=None, **extra):
        super(MetricReportField, self).__init__(
            behavior="metric",
            name=name,
            target_name=target_name,
            **extra
        )



import setuptools
from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='omni_reports',
    packages=setuptools.find_packages(),
    version='0.0.6',
    description='Omni Report Definition',
    author='Pareto Group',
    author_email='noreply@paretogroup.com.br',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/paretogroup/omni-reports',
    keywords=['pareto', 'api', 'reports'],
    classifiers=[],
)
